# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from hypernets.searchers.mcts_searcher import *
from hypernets.core.searcher import OptimizeDirection
from hypernets.core.ops import *
from hypernets.core.search_space import *

from hypernets.core.meta_learner import MetaLearner
from hypernets.core.trial import TrialHistory, DiskTrialStore, Trial
from hypernets.tests import test_output_dir

import numpy as np


class Test_MCTS():
    def get_space(self):
        space = HyperSpace()
        with space.as_default():
            p1 = Int(1, 100)
            p2 = Choice(['a', 'b'])
            p3 = Bool()
            p4 = Real(0.0, 1.0)
            id1 = Identity(p1=p1)
            id2 = Identity(p2=p2)(id1)
            id3 = Identity(p3=p3)(id2)
            id4 = Identity(p4=p4)(id3)
        return space

    def test_mctree(self):
        tree = MCTree(self.get_space, policy=UCT(), max_node_space=2)
        space_sample, node = tree.selection_and_expansion()
        assert node.param_sample.label == 'Param_Int_1-1-100-1'
        assert space_sample.all_assigned == False

        tree.back_propagation(node, np.random.uniform(0.1, 0.9))
        assert node.visits == 1
        assert tree.root.visits == 1

        space_sample, node = tree.selection_and_expansion()
        assert node.param_sample.label == 'Param_Int_1-1-100-1'
        assert space_sample.all_assigned == False

        tree.back_propagation(node, np.random.uniform(0.1, 0.9))
        assert node.visits == 1
        assert tree.root.visits == 2

        space_sample, node = tree.selection_and_expansion()
        assert node.param_sample.label == 'Param_Choice_1-[\'a\', \'b\']'
        tree.back_propagation(node, np.random.uniform(0.1, 0.9))
        assert node.visits == 1
        assert tree.root.visits == 3
        assert space_sample.all_assigned == False

        space_sample = tree.roll_out(space_sample, node)
        assert space_sample.all_assigned == True

        space_sample, node = tree.selection_and_expansion()
        assert node.param_sample.label == 'Param_Choice_1-[\'a\', \'b\']'
        tree.back_propagation(node, np.random.uniform(0.1, 0.9))
        assert node.visits == 1
        assert tree.root.visits == 4
        assert space_sample.all_assigned == False

        space_sample = tree.roll_out(space_sample, node)
        assert space_sample.all_assigned == True

        tree_str = str(tree.root)
        assert tree_str

    def test_mcts_searcher(self):
        searcher = MCTSSearcher(self.get_space, max_node_space=10)

        for i in range(1000):
            space_sample = searcher.sample()
            assert space_sample.all_assigned == True
            print(space_sample.params_summary())
            searcher.update_result(space_sample, np.random.uniform(0.1, 0.9))

        assert searcher.tree.root.visits == 1000
        assert len(searcher.nodes_map.items()) == 1000

    def test_mcts_searcher_parallelize(self):
        searcher = MCTSSearcher(self.get_space, max_node_space=10)
        history = TrialHistory(OptimizeDirection.Maximize)
        disk_trial_store = DiskTrialStore(f'{test_output_dir}/trial_store')
        meta_learner = MetaLearner(history, 'test_mcts_searcher_parallelize', disk_trial_store)
        searcher.set_meta_learner(meta_learner)
        trial_no = 1
        running_samples = []
        for i in range(10):
            space_sample = searcher.sample()
            running_samples.append(space_sample)
        assert searcher.tree.root.visits == 10
        for sample in running_samples:
            reward = np.random.uniform(0.1, 0.9)
            trial = Trial(sample, trial_no, reward, 10)
            history.append(trial)
            searcher.update_result(sample, reward)
            trial_no += 1

        assert searcher.tree.root.visits == 10
        assert len(searcher.nodes_map.items()) == 10
        running_samples = []
        for i in range(10):
            space_sample = searcher.sample()
            running_samples.append(space_sample)
        assert searcher.tree.root.visits == 20

        for sample in running_samples:
            reward = np.random.uniform(0.1, 0.9)
            trial = Trial(sample, trial_no, reward, 10)
            history.append(trial)
            searcher.update_result(sample, reward)

        assert searcher.tree.root.visits == 20

    def test_set_random_state(self):
        from hypernets.core import set_random_state
        set_random_state(9527)

        searcher = MCTSSearcher(self.get_space, max_node_space=10)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[98, 0, 1, 0.86], [2, 0, 1, 0.58], [2, 0, 1, 0.73], [2, 0, 1, 0.58], [2, 1, 0, 0.67],
                           [2, 0, 1, 0.01], [2, 0, 1, 0.22], [2, 0, 1, 0.55], [2, 0, 1, 0.32]]

        set_random_state(None)
        searcher = MCTSSearcher(self.get_space, max_node_space=10)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors != [[98, 0, 1, 0.86], [2, 0, 1, 0.58], [2, 0, 1, 0.73], [2, 0, 1, 0.58], [2, 1, 0, 0.67],
                           [2, 0, 1, 0.01], [2, 0, 1, 0.22], [2, 0, 1, 0.55], [2, 0, 1, 0.32]]

        set_random_state(9527)
        searcher = MCTSSearcher(self.get_space, max_node_space=10)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[98, 0, 1, 0.86], [2, 0, 1, 0.58], [2, 0, 1, 0.73], [2, 0, 1, 0.58], [2, 1, 0, 0.67],
                           [2, 0, 1, 0.01], [2, 0, 1, 0.22], [2, 0, 1, 0.55], [2, 0, 1, 0.32]]

        set_random_state(1)
        searcher = MCTSSearcher(self.get_space, max_node_space=10)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[73, 1, 0, 0.4], [2, 0, 0, 0.42], [2, 0, 0, 0.2], [2, 0, 0, 0.03], [2, 0, 0, 0.42],
                           [2, 1, 1, 0.14], [2, 0, 0, 0.8], [2, 1, 1, 0.31], [2, 0, 1, 0.88]]

        set_random_state(None)

    # def test_searcher_with_hp(self):
    #     def get_space():
    #         space = HyperSpace()
    #         with space.as_default():
    #             in1 = Input(shape=(10,))
    #             dense1 = Dense(10, activation=Choice(['relu', 'tanh', None]), use_bias=Bool())(in1)
    #             bn1 = BatchNormalization()(dense1)
    #             dropout1 = Dropout(Choice([0.3, 0.4, 0.5]))(bn1)
    #             output = Dense(2, activation='softmax', use_bias=True)(dropout1)
    #         return space
    #
    #     mcts = MCTSSearcher(get_space)
    #     hk = HyperKeras(mcts, optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'],
    #                     callbacks=[SummaryCallback()])
    #
    #     x = np.random.randint(0, 10000, size=(100, 10))
    #     y = np.random.randint(0, 2, size=(100), dtype='int')
    #
    #     hk.search(x, y, x, y, max_trials=10)
    #     assert hk.get_best_trial()
    #     best_trial = hk.get_best_trial()
    #
    #     estimator = hk.final_train(best_trial.space_sample, x, y)
    #     score = estimator.predict(x)
    #     result = estimator.evaluate(x, y)
    #     assert len(score) == 100
    #     assert result



# -*- coding:utf-8 -*-
import copy
import pickle
import time

from hypernets.core.dispatcher import Dispatcher
from hypernets.utils import logging, fs
from .grpc.search_driver_client import SearchDriverClient

logger = logging.get_logger(__name__)

_search_counter = 0


class ExecutorDispatcher(Dispatcher):
    def __init__(self, driver_address):
        super(ExecutorDispatcher, self).__init__()
        self.driver_address = driver_address

    def dispatch(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                 **fit_kwargs):

        if 'search_id' in fit_kwargs:
            search_id = fit_kwargs.pop('search_id')
        else:
            global _search_counter
            _search_counter += 1
            search_id = 'search-%02d' % _search_counter

        if logger.is_info_enabled():
            logger.info(f'[{search_id}] started')
            print(f'[{search_id}] connect to driver {self.driver_address}', end='')
        client = SearchDriverClient(self.driver_address, search_id)
        client.ping(wait=True)

        def response(item_x, success, reward=0.0, message=''):
            res = copy.copy(item_x)
            res.success = success
            res.reward = reward
            res.message = message
            return res

        trial_no = 0
        sch = client.search(search_id)
        try:
            item = next(sch)
            while item:
                if item.is_waiting():
                    if logger.is_info_enabled():
                        logger.info(f'[{search_id}] not found search, wait and continue')
                    time.sleep(1)
                    item = sch.send(response(item, True))
                    continue

                if item.is_finished():
                    if logger.is_info_enabled():
                        logger.info(f'[{search_id}] search finished, exit.')
                    # sch.send(None)
                    break

                if not item.is_ok():
                    if logger.is_info_enabled():
                        logger.info(f'[{search_id}] dispatched with {item.code}, exit.')
                    # sch.send(None)
                    break

                trial_no = item.trial_no if item.trial_no is not None else trial_no + 1
                detail = f'trial_no={trial_no}, space_id={item.space_id}, space_file={item.space_file}'
                if logger.is_info_enabled():
                    logger.info(f'[{search_id}] new trial:' + detail)
                try:
                    with fs.open(item.space_file, 'rb') as f:
                        space_sample = pickle.load(f)

                    for callback in hyper_model.callbacks:
                        # callback.on_build_estimator(hyper_model, space_sample, estimator, trial_no)
                        callback.on_trial_begin(hyper_model, space_sample, trial_no)

                    model_file = item.model_file
                    trial = hyper_model._run_trial(space_sample, trial_no, X, y, X_eval, y_eval, cv, num_folds,
                                                   model_file, **fit_kwargs)
                    if trial.reward != 0:
                        improved = hyper_model.history.append(trial)
                        for callback in hyper_model.callbacks:
                            callback.on_trial_end(hyper_model, space_sample, trial_no, trial.reward,
                                                  improved, trial.elapsed)
                    else:
                        for callback in hyper_model.callbacks:
                            callback.on_trial_error(hyper_model, space_sample, trial_no)

                    if trial_store is not None:
                        trial_store.put(dataset_id, trial)

                    item = sch.send(response(item, trial.reward != 0.0, trial.reward))
                except StopIteration:
                    break
                except KeyboardInterrupt:
                    if logger.is_info_enabled():
                        logger.info('KeyboardInterrupt')
                    break
                except Exception as e:
                    import traceback
                    msg = f'[{search_id}] {e.__class__.__name__}: {e}'
                    logger.error(msg + '\n' + traceback.format_exc())
                    item = sch.send(response(item, False, 0.0, msg))
        except StopIteration as e:
            pass
        finally:
            sch.close()
            client.close()

        if logger.is_info_enabled():
            logger.info(f'[{search_id}] search done, last trial_no={trial_no}')
        return trial_no



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from hypernets.discriminators import PercentileDiscriminator, ProgressivePercentileDiscriminator

from . import history, group_id


class Test_PercentileDiscriminator():
    def test_percentile(self):
        d = PercentileDiscriminator(50, min_trials=5, min_steps=5, stride=1, history=history, optimize_direction='min')
        p1 = d.is_promising([0.9, 0.9, 0.9, 0.9], group_id)
        assert p1 == True

        p2 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.9], group_id)
        assert p2 == False

        p2 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.525], group_id)
        assert p2 == False

        p2 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.524], group_id)
        assert p2 == True

        d = PercentileDiscriminator(0, min_trials=5, min_steps=5, stride=1, history=history, optimize_direction='min')
        p1 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.50], group_id)
        assert p1 == True
        p1 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.56], group_id)
        assert p1 == False

        d = PercentileDiscriminator(100, min_trials=5, min_steps=5, stride=1, history=history, optimize_direction='min')
        p1 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.55], group_id)
        assert p1 == False
        p1 = d.is_promising([0.9, 0.9, 0.9, 0.9, 0.49], group_id)
        assert p1 == True

    def test_progressive_percentile(self):
        d = ProgressivePercentileDiscriminator([100, 90, 80, 60, 50, 40, 30, 20, 10, 0], min_trials=5, min_steps=5,
                                               stride=1,
                                               history=history, optimize_direction='min')
        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1], group_id)
        assert p1 == True

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.56], group_id)
        assert p1 == False

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1, 0.1], group_id)
        assert p1 == True

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1, 0.45], group_id)
        assert p1 == False

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1], group_id)
        assert p1 == True

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.345], group_id)
        assert p1 == False

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1], group_id)
        assert p1 == True

        p1 = d.is_promising([0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.34], group_id)
        assert p1 == False



import re

import featuretools as ft
import numpy as np
from featuretools import variable_types
from sklearn.base import BaseEstimator, TransformerMixin

from hypernets.tabular.column_selector import column_all_datetime, column_number_exclude_timedelta
from hypernets.tabular.sklearn_ex import FeatureSelectionTransformer
from ._primitives import CrossCategorical, GeoHashPrimitive, DaskCompatibleHaversine, TfidfPrimitive

_named_primitives = [CrossCategorical, GeoHashPrimitive, DaskCompatibleHaversine, TfidfPrimitive]

_DEFAULT_PRIMITIVES_UNKNOWN = []
_DEFAULT_PRIMITIVES_NUMERIC = []
_DEFAULT_PRIMITIVES_CATEGORY = [CrossCategorical.name]
_DEFAULT_PRIMITIVES_DATETIME = ["month", "week", "day", "hour", "minute", "second", "weekday", "is_weekend"]
# _DEFAULT_PRIMITIVES_LATLONG = [GeoHashPrimitive.name, "haversine"]
_DEFAULT_PRIMITIVES_LATLONG = [GeoHashPrimitive.name, DaskCompatibleHaversine.name]
_DEFAULT_PRIMITIVES_TEXT = [TfidfPrimitive.name]

_named_primitives = {p.name: p for p in _named_primitives}

# _fix_feature_name
_pattern_to_sub = re.compile(r'[=()\[\], ]')
_chars_to_replace = {'+': 'A',
                     '-': 'S',
                     '*': 'X',
                     '/': 'D',
                     '%': 'M'}


def _fix_feature_name(s):
    s = _pattern_to_sub.sub('__', s)
    for k, v in _chars_to_replace.items():
        s = s.replace(k, v)
    return s


class FeatureGenerationTransformer(BaseEstimator, TransformerMixin):
    ft_index = 'e_hypernets_ft_index'

    def __init__(self, task=None, trans_primitives=None,
                 fix_input=False,
                 categories_cols=None,
                 continuous_cols=None,
                 datetime_cols=None,
                 latlong_cols=None,
                 text_cols=None,
                 max_depth=1,
                 max_features=-1,
                 drop_cols=None,
                 feature_selection_args=None,
                 fix_feature_names=True):
        """

        Args:
            trans_primitives:
                for categories: "cross_categorical"
                for continuous: "add_numeric","subtract_numeric","divide_numeric","multiply_numeric","negate","modulo_numeric","modulo_by_feature","cum_mean","cum_sum","cum_min","cum_max","percentile","absolute"
                for datetime: "year", "month", "week", "day", "hour", "minute", "second", "weekday", "is_weekend"
                for lat_long: "haversine", "geohash"
                for text: "num_characters", "num_words" + "tfidf"
            max_depth:
        """
        assert trans_primitives is None or isinstance(trans_primitives, (list, tuple)) and len(trans_primitives) > 0
        assert all([c is None or isinstance(c, (tuple, list))
                    for c in (categories_cols, continuous_cols, datetime_cols, latlong_cols, text_cols)])

        self.trans_primitives = trans_primitives
        self.max_depth = max_depth
        self.max_features = max_features
        self.task = task

        self.fix_input = fix_input
        self.categories_cols = categories_cols
        self.continuous_cols = continuous_cols
        self.datetime_cols = datetime_cols
        self.latlong_cols = latlong_cols
        self.text_cols = text_cols
        self.drop_cols = drop_cols
        self.feature_selection_args = feature_selection_args
        self.fix_feature_names = fix_feature_names

        # fitted
        self.original_cols = []
        self.selection_transformer = None
        self.imputed_input_ = None
        self.feature_defs_ = None
        self.transformed_feature_names_ = None
        self.feature_defs_names_ = None

    def fit(self, X, y=None, **kwargs):
        original_cols = X.columns.to_list()

        if self.feature_selection_args is not None:
            assert y is not None, '`y` must be provided for feature selection.'
            self.feature_selection_args['reserved_cols'] = original_cols
            self.selection_transformer = FeatureSelectionTransformer(task=self.task, **self.feature_selection_args)

        if self.categories_cols is None:
            self.categories_cols = []
        if self.continuous_cols is None:
            self.continuous_cols = column_number_exclude_timedelta(X)
        if self.datetime_cols is None:
            self.datetime_cols = column_all_datetime(X)
        if self.latlong_cols is None:
            self.latlong_cols = []
        if self.text_cols is None:
            self.text_cols = []

        if self.fix_input:
            self.imputed_input_ = self._detect_impute_dict(X)
            X = self._replace_invalid_values(X, self.imputed_input_)

        if self.trans_primitives is None:
            self.trans_primitives = self._default_trans_primitives(X, y)

        trans_primitives = self.trans_primitives
        if any([isinstance(p, str) and p in _named_primitives.keys() for p in trans_primitives]):
            trans_primitives = [_named_primitives.get(p, p) if isinstance(p, str) else p
                                for p in trans_primitives]

        es = ft.EntitySet(id='es_hypernets_fit')
        make_index = self.ft_index not in original_cols
        feature_type_dict = self._get_feature_types(X)
        es.entity_from_dataframe(entity_id='e_hypernets_ft', dataframe=X, variable_types=feature_type_dict,
                                 make_index=make_index, index=self.ft_index)
        feature_matrix, feature_defs = ft.dfs(entityset=es, target_entity="e_hypernets_ft",
                                              ignore_variables={"e_hypernets_ft": []},
                                              return_variable_types="all",
                                              trans_primitives=trans_primitives,
                                              drop_exact=self.drop_cols,
                                              max_depth=self.max_depth,
                                              max_features=self.max_features,
                                              features_only=False)
        X.pop(self.ft_index)

        self.feature_defs_ = feature_defs
        self.original_cols = original_cols
        feature_defs_names, replaced_feature_defs_names, = self._get_transformed_feature_names(feature_defs)
        self.transformed_feature_names_ = replaced_feature_defs_names
        self.feature_defs_names_ = feature_defs_names

        if self.selection_transformer is not None:
            self.selection_transformer.fit(feature_matrix, y)
            selected_defs = []
            for fea in self.feature_defs_:
                if fea._name in self.selection_transformer.columns_:
                    selected_defs.append(fea)
            self.feature_defs_ = selected_defs

        return self

    def transform(self, X, y=None):
        # 1. check is fitted and values
        assert self.feature_defs_ is not None, 'Please fit it first.'

        # 2. fix input
        if self.fix_input:
            X = self._replace_invalid_values(X, self.imputed_input_)

        # 3. transform
        es = ft.EntitySet(id='es_hypernets_transform')
        feature_type_dict = self._get_feature_types(X)
        make_index = self.ft_index not in X.columns.to_list()
        es.entity_from_dataframe(entity_id='e_hypernets_ft', dataframe=X, variable_types=feature_type_dict,
                                 make_index=make_index, index=self.ft_index)
        Xt = ft.calculate_feature_matrix(self.feature_defs_, entityset=es, n_jobs=1, verbose=False)
        if make_index:
            X.pop(self.ft_index)
            if self.ft_index in Xt.columns.to_list():
                Xt.pop(self.ft_index)
        Xt = Xt.replace([np.inf, -np.inf], np.nan)

        if self.fix_feature_names:
            Xt = self._fix_transformed_feature_names(Xt)

        return Xt

    @staticmethod
    def _merge_dict(dest_dict, *dicts):
        for d in dicts:
            for k, v in d.items():
                dest_dict.setdefault(k, v)

    @property
    def classes_(self):
        if self.feature_defs_ is None:
            return None
        feats = [fea._name for fea in self.feature_defs_]
        return feats

    def _default_trans_primitives(self, X, y):
        primitives = []

        if self.categories_cols:
            primitives += _DEFAULT_PRIMITIVES_CATEGORY
        if self.continuous_cols:
            primitives += _DEFAULT_PRIMITIVES_NUMERIC
        if self.datetime_cols:
            primitives += _DEFAULT_PRIMITIVES_DATETIME
        if self.text_cols:
            primitives += _DEFAULT_PRIMITIVES_TEXT
        if self.latlong_cols:
            primitives += _DEFAULT_PRIMITIVES_LATLONG

        return primitives

    def _detect_impute_dict(self, X):
        imputed_input = {}
        if len(self.continuous_cols) > 0:
            _mean = X[self.continuous_cols].mean()
            if hasattr(_mean, 'compute'):
                _mean = _mean.compute()
            self._merge_dict(imputed_input, _mean.to_dict())
        if len(self.datetime_cols) > 0:
            _mode = X[self.datetime_cols].mode()
            if hasattr(_mode, 'compute'):
                _mode = _mode.compute()
            self._merge_dict(imputed_input, _mode.iloc[0].to_dict())
        if len(self.text_cols) > 0:
            self._merge_dict(imputed_input, {c: '' for c in self.text_cols})
        if len(self.latlong_cols) > 0:
            pass  # do nothing

        return imputed_input

    def _replace_invalid_values(self, df, imputed_dict):
        df = df.replace([np.inf, -np.inf], np.nan)
        if imputed_dict is not None and len(imputed_dict) > 0:
            df = df.fillna(imputed_dict)
        else:
            df = df.fillna(0)

        return df

    def _get_transformed_feature_names(self, feature_defs):
        names = [n for f in feature_defs for n in f.get_feature_names()]
        if self.fix_feature_names:
            replaced_names = [_fix_feature_name(n) for n in names]
        else:
            replaced_names = names

        return names, replaced_names

    def _fix_transformed_feature_names(self, df):
        if self.fix_feature_names and hasattr(df, 'columns'):
            columns = [_fix_feature_name(n) for n in df.columns.to_list()]
            df.columns = columns

        return df

    def _get_feature_types(self, X):
        feature_types = {}
        original_cols = X.columns.to_list()

        known_cols = self.categories_cols + self.continuous_cols + self.datetime_cols \
                     + self.latlong_cols + self.text_cols
        unknown_cols = [c for c in original_cols if c not in known_cols]
        self._merge_dict(feature_types,
                         {c: variable_types.Numeric for c in self.continuous_cols},
                         {c: variable_types.Datetime for c in self.datetime_cols},
                         {c: variable_types.LatLong for c in self.latlong_cols},
                         {c: variable_types.NaturalLanguage for c in self.text_cols},
                         {c: variable_types.Categorical for c in self.categories_cols},
                         {c: variable_types.Unknown for c in unknown_cols},
                         )
        return feature_types



# -*- coding:utf-8 -*-
"""

"""

import json
import os
import sys
import tempfile

import fsspec

from hypernets.conf import Configurable, configure, Unicode
from . import logging

logger = logging.get_logger(__name__)
is_windows = sys.platform.find('win') >= 0


@configure()
class StorageCfg(Configurable):
    kind = Unicode('file',
                   help='storage kind (fsspec protocol). "file" and "s3" were tested.'
                   ).tag(config=True)
    root = Unicode('',
                   help='storage root path.'
                   ).tag(config=True)
    options = Unicode('',
                      help='storage options, json string. see fsspec from more details.'
                      ).tag(config=True)
    local_root = Unicode(os.path.join(tempfile.gettempdir(), 'cache'),
                         help='local root path, used as temp.'
                         ).tag(config=True)


class FileSystemAdapter(object):
    def __init__(self, remote_root, local_root, remote_sep):
        super(FileSystemAdapter, self).__init__()

        # self.fs = fs
        self.remote_root = remote_root
        self.local_root = local_root
        self.remote_sep = remote_sep
        self.remote_root_alias = remote_root.replace('\\', '/') if is_windows else None

    def _inner_to_rpath(self, rpath):
        if rpath.startswith(self.remote_root):
            return rpath

        if self.remote_root_alias and rpath.startswith(self.remote_root_alias):
            return rpath

        if is_windows and rpath.find(':') > 0:
            return rpath

        return self.remote_root.rstrip(self.remote_sep) + self.remote_sep + rpath.lstrip(self.remote_sep)

    def to_rpath(self, rpath):
        # return os.path.join(remote_root, rpath)
        assert rpath

        if isinstance(rpath, str):
            return self._inner_to_rpath(rpath)
        elif isinstance(rpath, (list, tuple)):
            return [self._inner_to_rpath(p) for p in rpath]
        else:
            logger.warn(f'Unexpected rpath type: {type(rpath)}, rpath: {rpath}')
            return rpath

    def to_lpath(self, lpath):
        assert lpath
        if lpath.startswith(self.local_root):
            return lpath

        return os.path.join(self.local_root, lpath)

    def strip_rpath(self, rpath, align_with):
        if align_with.startswith(self.remote_root):
            # if align_with.startswith(remote_sep) and rpath.startswith(align_with.lstrip(remote_sep)):
            #     rpath = remote_sep + rpath
            return rpath

        n = rpath.find(self.to_rpath(align_with).lstrip(self.remote_sep))
        if n == 0 or n == 1:
            rpath = rpath[len(self.remote_root.lstrip(self.remote_sep)):]
            rpath = rpath[rpath.find(align_with):]

        return rpath

    def handle_find(self, result, rpath, *args, **kwargs):
        if kwargs.get('detail'):
            if isinstance(result, dict):
                # fixed = dict([(k, {**v, 'name': strip_rpath(v['name'], rpath)}
                #                ) for k, v in result.items()])
                fixed = {}
                for k, v in result.items():
                    key = self.strip_rpath(k, rpath)
                    value = {**v, 'name': self.strip_rpath(v['name'], rpath)}
                    if 'Key' in v.keys():
                        value['Key'] = self.strip_rpath(v['Key'], rpath)
                    fixed[key] = value
            elif isinstance(result, (list, tuple)):
                fixed = [{**v, 'name': self.strip_rpath(v['name'], rpath)}
                         for v in result]
            else:
                assert False, f'Unexpected result type: {type(result)}'
        else:
            assert isinstance(result, (list, tuple))
            fixed = [self.strip_rpath(path, rpath) for path in result]
        return fixed

    def handle_glob(self, result, rpath, *args, **kwargs):
        assert isinstance(result, (list, tuple))
        n = rpath.find('*')
        if n >= 0:
            fixed = [self.strip_rpath(path, rpath[:n]) for path in result]
        else:
            fixed = [self.strip_rpath(path, rpath) for path in result]
        return fixed

    def handle_info(self, result, rpath, *args, **kwargs):
        assert isinstance(result, dict)
        fixed = {**result, 'name': self.strip_rpath(result['name'], rpath)}
        return fixed

    def handle_walk(self, result, rpath, *args, **kwargs):
        assert type(result).__name__ == 'generator'
        for root, dirs, files in result:
            yield self.strip_rpath(root, rpath), dirs, files

    def fix_r(self, fn, post_handler):
        def execute(rpath, *args, **kwargs):
            result = fn(self.to_rpath(rpath), *args, **kwargs)
            if post_handler:
                result = post_handler(result, rpath, *args, **kwargs)

            # print('-' * 20, fn.__name__, rpath, args, kwargs, result)
            return result

        return execute

    def fix_rr(self, fn, post_handler):
        def execute(rpath1, rpath2, *args, **kwargs):
            return fn(self.to_rpath(rpath1),
                      self.to_rpath(rpath2),
                      *args, **kwargs)

        return execute

    def fix_lr(self, fn, post_handler):
        def execute(lpath, rpath, *args, **kwargs):
            return fn(self.to_lpath(lpath),
                      self.to_rpath(rpath),
                      *args, **kwargs)

        return execute

    def fix_rl(self, fn, post_handler):
        def execute(rpath, lpath, *args, **kwargs):
            return fn(self.to_rpath(rpath),
                      self.to_lpath(lpath),
                      *args, **kwargs)

        return execute

    # functions with one remote path only
    @property
    def fn_r(self):
        return ['cat', 'cat_file', 'checksum', 'created',
                'delete', 'disk_usage', 'du',
                'exists', 'expand_path', 'find', 'glob', 'head',
                'info', 'invalidate_cache', 'isdir', 'isfile',
                'listdir', 'ls',
                'makedir', 'makedirs', 'mkdir', 'mkdirs', 'modified',
                'open', 'pipe', 'pipe_file',
                'rm', 'rm_file', 'rmdir',
                'sign', 'size', 'stat',
                'tail', 'touch', 'ukey', 'walk']

    # functions with remote-remote path pair
    @property
    def fn_rr(self):
        return ['copy', 'cp', 'move', 'mv', 'rename', ]

    # functions with remote-local path pair
    @property
    def fn_rl(self):
        return ['download', 'get', 'get_file', ]

    # functions with local-remote path pair
    @property
    def fn_lr(self):
        return ['put', 'put_file', 'upload']

    # functions without any path
    @property
    def fn_unhandled(self):
        return ['clear_instance_cache', 'current',
                'end_transaction', 'from_json', 'get_mapper',
                'read_block', 'start_transaction', 'to_json']

    # functions to handle result
    @property
    def fn_post_process(self):
        return {'ls': self.handle_find,
                'find': self.handle_find,
                'glob': self.handle_glob,
                'info': self.handle_info,
                'walk': self.handle_walk,
                }

    @property
    def fn_fix_pairs(self):
        return [(self.fn_r, self.fix_r),
                (self.fn_rr, self.fix_rr),
                (self.fn_lr, self.fix_lr),
                (self.fn_rl, self.fix_rl)]

    def __call__(self, fs, *args, **kwargs):
        assert not hasattr(fs, '_hyn_adapted_')

        # decorate listed functions
        post_processes = self.fn_post_process
        for fns, fix in self.fn_fix_pairs:
            for fn in fns:
                # assert hasattr(fs, fn), f'fn:{fn}'
                if not hasattr(fs, fn):
                    continue

                original_fn = f'_orig_{fn}_'
                assert not hasattr(fs, original_fn)

                f = getattr(fs, fn)
                setattr(fs, original_fn, f)
                setattr(fs, fn, fix(f, post_processes.get(fn)))

        # decorate '__reduce__' (pickle compatible)
        f = getattr(fs, '__reduce__')
        setattr(fs, '_orig__reduce__', f)
        setattr(fs, '__reduce__', _fs_reduce)

        # mark adapted
        setattr(fs, '_hyn_adapted_', 1)

        return fs


class S3FileSystemAdapter(FileSystemAdapter):
    def __init__(self, remote_root, local_root, remote_sep):
        remote_root = remote_root.lstrip(remote_sep)

        super().__init__(remote_root, local_root, remote_sep)

    def handle_private_ls(self, result, rpath, *args, **kwargs):
        assert isinstance(result, (list, tuple))
        fixed = [{**v,
                  'name': self.strip_rpath(v['name'], rpath),
                  'Key': self.strip_rpath(v['Key'], rpath),
                  }
                 for v in result]
        return fixed

    @property
    def fn_r(self):
        from distutils.version import LooseVersion
        import s3fs
        ver = LooseVersion(s3fs.__version__)

        if ver < LooseVersion('0.5.0'):
            return super().fn_r + ['_ls']
        else:
            return super().fn_r

    @property
    def fn_post_process(self):
        return {**super().fn_post_process, '_ls': self.handle_private_ls}


def get_filesystem(fs_type, fs_root, fs_options) -> fsspec.AbstractFileSystem:
    if fs_options is None or len(fs_options) == 0:
        fs = fsspec.filesystem(fs_type, skip_instance_cache=True)
    else:
        try:
            parsed = json.loads(fs_options)
        except json.JSONDecodeError as e:
            msg = 'Failed to parse storage options as json, ' \
                  f'current settings:\n{fs_options}'
            raise Exception(msg) from e

        if not isinstance(parsed, dict):
            msg = 'Storage options should be json dictionary, ' \
                  f'current settings:\n{fs_options}'
            raise Exception(msg)

        fs = fsspec.filesystem(fs_type, skip_instance_cache=True, **parsed)

    if type(fs).__name__.lower().find('local') >= 0:
        if fs_root is None or fs_root == '':
            fs_root = os.path.join(tempfile.gettempdir(), 'workdir')
            # logger.info(f'use {fs_root} as working directory.')

        remote_root = os.path.abspath(os.path.expanduser(fs_root))
        try:
            if not fs.exists(remote_root):
                fs.mkdirs(remote_root, exist_ok=True)
            else:
                with tempfile.TemporaryFile(prefix='hyn_', dir=remote_root) as t:
                    t.write(b'.')
        except PermissionError as e:
            logger.warn(f'{type(e).__name__}: working directory "{remote_root}"')
            logger.warn(e)

        local_root = remote_root
        is_local = True
    else:
        remote_root = fs_root if fs_root else '/tmp'
        if not fs.exists(remote_root):
            fs.mkdirs(remote_root, exist_ok=True)

        local_root = StorageCfg.local_root
        local_root = os.path.abspath(os.path.expanduser(local_root))
        os.makedirs(local_root, exist_ok=True)
        is_local = False

    # return fix_filesystem(fs, remote_root, local_root)
    if type(fs).__name__ == 'S3FileSystem':
        return S3FileSystemAdapter(remote_root, local_root, fs.sep)(fs)
    else:
        return FileSystemAdapter(remote_root, local_root, os.path.sep if is_local else fs.sep)(fs)


def _fs_reduce(*args):
    return get_filesystem, (StorageCfg.kind if len(StorageCfg.kind) > 0 else None,
                            StorageCfg.root if len(StorageCfg.root) > 0 else None,
                            StorageCfg.options if len(StorageCfg.options) > 0 else None)


filesystem = _fs_reduce()[0](*_fs_reduce()[1])



import queue
import sys

import grpc

from hypernets.dispatchers.process.grpc.proto import proc_pb2_grpc
from hypernets.dispatchers.process.grpc.proto.proc_pb2 import DataChunk, ProcessRequest, DownloadRequest
from hypernets.utils import logging

logger = logging.get_logger(__name__)


class ProcessBrokerClient(object):

    def __init__(self, server):
        super(ProcessBrokerClient, self).__init__()
        self.channel = grpc.insecure_channel(server)
        self.stub = proc_pb2_grpc.ProcessBrokerStub(self.channel)

        self.server = server

    def __del__(self):
        self.channel.close()

    def run(self, args, program=None, buffer_size=-1, cwd=None, stdout=None, stderr=None):
        if stdout is None:
            stdout = sys.stdout
        if stderr is None:
            stdout = sys.stderr

        encoding = getattr(stdout, 'encoding', None)

        result = {'running': True}
        ack_queue = queue.Queue()

        if encoding:
            handles = {DataChunk.OUT: lambda x: stdout.write(x.decode(encoding)),
                       DataChunk.DATA: lambda x: stdout.write(x.decode(encoding)),
                       DataChunk.ERR: lambda x: stderr.write(x.decode(encoding)),
                       DataChunk.END: lambda x: result.update({'code': x, 'running': False}),
                       }
        else:
            handles = {DataChunk.OUT: lambda x: stdout.write(x),
                       DataChunk.DATA: lambda x: stdout.write(x),
                       DataChunk.ERR: lambda x: stderr.write(x),
                       DataChunk.END: lambda x: result.update({'code': x, 'running': False}),
                       }

        def fire_request():
            request = ProcessRequest(args=args,
                                     program=program if program else b'',
                                     buffer_size=buffer_size,
                                     encoding=encoding if encoding else b'',
                                     cwd=cwd if cwd else b'')
            yield request

            dummy = ProcessRequest()
            while result['running']:
                if ack_queue.get():
                    yield dummy

        try:
            response = self.stub.run(fire_request())
            for chunk in response:
                fn = handles.get(chunk.kind, None)
                if fn:
                    fn(chunk.data)
                else:
                    logger.warning(f'unexpected chunk kind: {chunk.kind}')
                ack = chunk.kind != DataChunk.END
                ack_queue.put(ack)
            if 'code' in result.keys():
                try:
                    return int(result['code'])
                except ValueError:
                    return 90
            else:
                return 0
        except grpc.RpcError as e:
            try:
                msg = f'RpcError {e.__class__.__name__}: {e.code()}'
                logger.error(msg)
            except Exception:
                import traceback
                msg = f'[GRPC {self.server}] {e.__class__.__name__}:\n'
                logger.error(msg + traceback.format_exc())
            return 99

    def download(self, remote_file_path, local_file_object, buffer_size=-1):
        encoding = getattr(local_file_object, 'encoding', None)

        def dump_with_encoding(data):
            local_file_object.write(data.decode(encoding))

        if encoding:
            fn = dump_with_encoding
        else:
            fn = local_file_object.write

        errs = []
        try:
            request = DownloadRequest(peer='hi',
                                      path=remote_file_path,
                                      encoding=encoding if encoding else '',
                                      buffer_size=buffer_size)
            response = self.stub.download(request)
            for chunk in response:
                if chunk.kind == DataChunk.DATA:
                    fn(chunk.data)
                elif chunk.kind == DataChunk.END:
                    pass  #
                else:
                    errs.append(f'[{chunk.kind}] ' + chunk.data.decode(encoding))

            if errs:
                raise Exception('\n'.join(errs))
        except grpc.RpcError as e:
            try:
                msg = f'RpcError {e.__class__.__name__}: {e.code()}'
                logger.error(msg)
            except Exception:
                import traceback
                msg = f'[GRPC {self.server}] {e.__class__.__name__}:\n'
                logger.error(msg + traceback.format_exc())

            raise Exception(msg)



import queue
import subprocess
import time
from threading import Thread

from grpc import RpcError

from hypernets.dispatchers.process.grpc.proto import proc_pb2_grpc
from hypernets.dispatchers.process.grpc.proto.proc_pb2 import DataChunk
from hypernets.utils import logging

logger = logging.get_logger(__name__)


class ProcessBrokerService(proc_pb2_grpc.ProcessBrokerServicer):
    def __init__(self):
        super(ProcessBrokerService, self).__init__()

    @staticmethod
    def _read_data(f, q, buffer_size, encoding, data_kind):

        try:
            data = f.read(buffer_size)
            while data and len(data) > 0:
                if encoding:
                    chunk = DataChunk(kind=data_kind, data=data.encode(encoding))
                else:
                    chunk = DataChunk(kind=data_kind, data=data)
                q.put(chunk)
                data = f.read(buffer_size)
        except ValueError as e:
            logger.error(e)

    def run(self, request_iterator, context):
        it = iter(request_iterator)
        request = next(it)

        program = request.program
        args = request.args
        cwd = request.cwd
        buffer_size = request.buffer_size
        encoding = request.encoding
        if encoding is None or len(encoding) == 0:
            encoding = None

        with subprocess.Popen(args, buffer_size,
                              program if len(program) > 0 else None,
                              cwd=cwd if len(cwd) > 0 else None,
                              stdin=None,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              encoding=encoding,
                              shell=False) as p:
            pid = p.pid
            start_at = time.time()
            peer = context.peer()
            if logger.is_info_enabled():
                logger.info(f'[{pid}] started, peer: {peer}, cmd:' + ' '.join(args), )

            data_queue = queue.Queue()
            t_out = Thread(target=self._read_data,
                           args=(p.stdout, data_queue, buffer_size, encoding, DataChunk.OUT))
            t_err = Thread(target=self._read_data,
                           args=(p.stderr, data_queue, buffer_size, encoding, DataChunk.ERR))
            t_out.start()
            t_err.start()

            # report pid to client
            yield DataChunk(kind=DataChunk.ERR, data=f'pid: {pid}\n'.encode())

            try:
                while next(it):
                    chunk = None
                    while context.is_active() and \
                            (t_out.is_alive() or t_err.is_alive() or not data_queue.empty()):
                        try:
                            chunk = data_queue.get(False)
                            yield chunk
                            break
                        except queue.Empty:
                            time.sleep(0.1)
                    if not context.is_active():
                        p.kill()
                        code = 'killed (peer shutdown)'
                        break
                    elif chunk is None:  # process exit and no more output
                        code = p.poll()
                        yield DataChunk(kind=DataChunk.END, data=str(code).encode())
                        # break
            except StopIteration as e:
                pass
            except RpcError as e:
                logger.error(e)
                code = 'rpc error'
            except Exception:
                import traceback
                traceback.print_exc()
                code = 'exception'

        if logger.is_info_enabled():
            logger.info('[%s] done with code %s, elapsed %.3f seconds.'
                        % (pid, code, time.time() - start_at))

    def download(self, request, context):
        try:
            peer = request.peer
            path = request.path
            encoding = request.encoding
            buffer_size = request.buffer_size
            if buffer_size is None or buffer_size <= 0:
                buffer_size = 4096

            # check peer here

            start_at = time.time()
            total = 0
            if encoding:
                with open(path, 'r', encoding=encoding) as f:
                    data = f.read(buffer_size)
                    while data and len(data) > 0:
                        if not context.is_active():
                            break
                        encoded_data = data.encode(encoding)
                        chunk = DataChunk(kind=DataChunk.DATA, data=encoded_data)
                        total += len(encoded_data)
                        yield chunk
                        data = f.read(buffer_size)
            else:
                with open(path, 'rb') as f:
                    data = f.read(buffer_size)
                    while data and len(data) > 0:
                        if not context.is_active():
                            break
                        chunk = DataChunk(kind=DataChunk.DATA, data=data)
                        total += len(data)
                        yield chunk
                        data = f.read(buffer_size)

            if not context.is_active():
                if logger.is_info_enabled():
                    logger.info('download %s broke (peer shutdown), %s bytes sent, elapsed %.3f seconds' %
                                (path, total, time.time() - start_at))
            else:
                yield DataChunk(kind=DataChunk.END, data=b'')
                if logger.is_info_enabled():
                    logger.info('download %s (%s bytes) in %.3f seconds, encoding=%s' %
                                (path, total, time.time() - start_at, encoding))
        except Exception as e:
            import traceback
            import sys
            msg = f'{e.__class__.__name__}:\n'
            msg += traceback.format_exc()
            logger.error(msg)
            yield DataChunk(kind=DataChunk.EXCEPTION, data=msg.encode())


def serve(addr, max_workers=10):
    import grpc
    from concurrent import futures

    if logger.is_info_enabled():
        logger.info(f'start broker at {addr}')
    service = ProcessBrokerService()
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
    proc_pb2_grpc.add_ProcessBrokerServicer_to_server(service, server)

    server.add_insecure_port(addr)
    server.start()

    return server, service



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import time
from sklearn.metrics import get_scorer
from sklearn.model_selection import train_test_split
from collections import defaultdict


class BaseEstimator():
    def __init__(self, task=None):
        self.task = task
        self.name = None

    def train(self, X, y, X_test):
        raise NotImplementedError

    def predict(self, X):
        raise NotImplementedError

    def predict_proba(self, X):
        raise NotImplementedError

    def proba2predict(self, proba, proba_threshold=0.5):
        if self.task == 'regression':
            return proba
        if proba.shape[-1] > 2:
            predict = proba.argmax(axis=-1)
        elif proba.shape[-1] == 2:
            predict = (proba[:, 1] > proba_threshold).astype('int32')
        else:
            predict = (proba > proba_threshold).astype('int32')
        return predict


class Evaluator():
    def evaluate(self, data, target, task, estimators, scorers, test_size=0.3, random_state=9527):
        if isinstance(data, tuple):
            assert len(data) == 2
            X_train = data[0]
            X_test = data[1]
            y_train = X_train.pop(target)
            y_test = X_test.pop(target)
        else:
            y = data.pop(target)
            if task == 'binary':
                stratify = y
            else:
                stratify = None
            X_train, X_test, y_train, y_test = train_test_split(data, y, test_size=test_size, random_state=random_state,
                                                                stratify=stratify)

        result = []
        errors = []
        for estimator in estimators:
            try:
                starttime = time.time()
                estimator.train(X_train, y_train, X_test)
                elapsed = time.time() - starttime
                for scoring in scorers:
                    if isinstance(scoring, str):
                        scorer = get_scorer(scoring)
                        metric = scoring
                    else:
                        metric = str(scoring)
                    score = scorer(estimator, X_test, y_test) * scorer._sign
                    result.append((estimator.name, metric, score, elapsed, time.strftime("%Y-%m-%d %H:%M:%S")))
            except Exception as e:
                errors.append((estimator.name, str(e)))
                import traceback
                print(traceback.format_exc())

        return result, errors



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import copy
import pickle

import numpy as np
from sklearn.model_selection import StratifiedKFold

from hypernets.utils import fs


class BaseEnsemble():
    def __init__(self, task, estimators, need_fit=False, n_folds=5, method='soft', random_state=9527):
        self.task = task
        self.estimators = list(estimators)
        self.need_fit = need_fit
        self.method = method
        self.n_folds = n_folds
        self.random_state = random_state
        self.classes_ = None
        self.preds_shape_ = None
        for est in estimators:
            if est is not None and self.classes_ is None and hasattr(est, 'classes_'):
                self.classes_ = est.classes_
                break

    def __predict(self, estimator, X):
        if self.task == 'regression':
            pred = estimator.predict(X)
        else:
            if self.classes_ is None and hasattr(estimator, 'classes_'):
                self.classes_ = estimator.classes_
            pred = estimator.predict_proba(X)
            if self.method == 'hard':
                pred = self.proba2predict(pred)
        return pred

    def proba2predict(self, proba, proba_threshold=0.5):
        assert len(proba.shape) <= 2
        if self.task == 'regression':
            return proba
        if len(proba.shape) == 2:
            if proba.shape[-1] > 2:
                predict = proba.argmax(axis=-1)
            else:
                predict = (proba[:, -1] > proba_threshold).astype('int32')
        else:
            predict = (proba > proba_threshold).astype('int32')

        return predict

    def fit(self, X, y, est_predictions=None):
        assert y is not None
        if est_predictions is not None:
            # print(f'est_predictions.shape:{est_predictions.shape}, estimators:{len(self.estimators)}')
            if self.task == 'regression' or self.method == 'hard':
                assert est_predictions.shape == (len(y), len(
                    self.estimators)), f'shape is not equal, may be a wrong task type. task:{self.task},  est_predictions.shape: {est_predictions.shape}, (len(y), len(self.estimators)):{(len(y), len(self.estimators))}'
            else:
                assert len(est_predictions.shape) == 3
                assert est_predictions.shape[0] == len(y)
                assert est_predictions.shape[1] == len(self.estimators)
        else:
            assert X is not None
            if self.task == 'regression' or self.method == 'hard':
                est_predictions = np.zeros((len(y), len(self.estimators)), dtype=np.float64)
            if self.need_fit:
                iterators = StratifiedKFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)
                for fold, (train, test) in enumerate(iterators.split(X, y)):
                    for n, estimator in enumerate(self.estimators):
                        X_train = X.iloc[train]
                        y_train = y.iloc[train]
                        X_test = X.iloc[test]
                        estimator.fit(X_train, y_train)
                        pred = self.__predict(estimator, X_test)
                        if est_predictions is None:
                            est_predictions = np.zeros((len(y), len(self.estimators), pred.shape[1]), dtype=np.float64)
                        est_predictions[test, n] = pred

            else:
                for n, estimator in enumerate(self.estimators):
                    pred = self.__predict(estimator, X)
                    if est_predictions is None:
                        est_predictions = np.zeros((len(y), len(self.estimators), pred.shape[1]), dtype=np.float64)
                    est_predictions[:, n] = pred

        self.preds_shape_ = est_predictions.shape
        self.fit_predictions(est_predictions, y)

    def X2predictions(self, X):
        if self.task == 'regression' or self.method == 'hard':
            est_predictions = np.zeros((len(X), len(self.estimators)), dtype=np.float64)
        else:
            est_predictions = np.zeros((len(X), len(self.estimators), self.preds_shape_[-1]), dtype=np.float64)
        for n, estimator in enumerate(self.estimators):
            if estimator is not None:
                pred = self.__predict(estimator, X)
                if self.task == 'regression' and len(pred.shape) > 1:
                    assert pred.shape[1] == 1
                    pred = pred.reshape(pred.shape[0])
                est_predictions[:, n] = pred
        return est_predictions

    def predict(self, X):
        est_predictions = self.X2predictions(X)
        pred = self.predictions2predict(est_predictions)
        if self.task != 'regression' and self.classes_ is not None:
            pred = np.array(self.classes_).take(pred, axis=0)
        return pred

    def predict_proba(self, X):
        est_predictions = self.X2predictions(X)
        return self.predictions2predict_proba(est_predictions)

    def fit_predictions(self, predictions, y_true):
        raise NotImplementedError()

    def predictions2predict_proba(self, predictions):
        raise NotImplementedError()

    def predictions2predict(self, predictions):
        raise NotImplementedError()

    def save(self, model_path):
        if not model_path.endswith(fs.sep):
            model_path = model_path + fs.sep
        if not fs.exists(model_path):
            fs.mkdirs(model_path, exist_ok=True)

        stub = copy.copy(self)
        estimators = self.estimators
        if estimators is not None:
            stub.estimators = [None for _ in estimators]  # keep size

        if estimators is not None:
            for i, est in enumerate(estimators):
                est_pkl = f'{model_path}{i}.pkl'
                est_model = f'{model_path}{i}.model'
                for t in [est_pkl, est_model]:
                    if fs.exists(t):
                        fs.rm(t)

                if est is None:
                    continue
                with fs.open(est_pkl, 'wb') as f:
                    pickle.dump(est, f, protocol=pickle.HIGHEST_PROTOCOL)

                if hasattr(est, 'save') and hasattr(est, 'load'):
                    est.save(est_model)

        with fs.open(f'{model_path}ensemble.pkl', 'wb') as f:
            pickle.dump(stub, f, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load(model_path):
        if not model_path.endswith(fs.sep):
            model_path = model_path + fs.sep

        with fs.open(f'{model_path}ensemble.pkl', 'rb') as f:
            stub = pickle.load(f)

        if stub.estimators is not None:
            for i in range(len(stub.estimators)):
                if fs.exists(f'{model_path}{i}.pkl'):
                    with fs.open(f'{model_path}{i}.pkl', 'rb') as f:
                        est = pickle.load(f)
                    if fs.exists(f'{model_path}{i}.model') and hasattr(est, 'load'):
                        est = est.load(f'{model_path}{i}.model')
                    stub.estimators[i] = est

        return stub



# -*- coding:utf-8 -*-
"""

"""
import dask
import numpy as np
import pandas as pd
from dask import dataframe as dd
from scipy.stats import skew, kurtosis
from sklearn.compose import make_column_selector

from .cfg import TabularCfg as cfg

try:
    import jieba

    _jieba_installed = True
except ImportError:
    _jieba_installed = False


class ColumnSelector(make_column_selector):
    __doc__ = make_column_selector.__doc__

    def __init__(self, pattern=None, *, dtype_include=None, dtype_exclude=None):
        super(ColumnSelector, self).__init__(pattern, dtype_include=dtype_include, dtype_exclude=dtype_exclude)

    def __call__(self, df):
        if isinstance(df, dd.DataFrame):
            # # if not hasattr(df, 'iloc'):
            # #     raise ValueError("make_column_selector can only be applied to "
            # #                      "pandas dataframes")
            # df_row = df.iloc[:1]
            df_row = df

            if self.dtype_include is not None or self.dtype_exclude is not None:
                df_row = df_row.select_dtypes(include=self.dtype_include, exclude=self.dtype_exclude)
            cols = df_row.columns
            if self.pattern is not None:
                cols = cols[cols.str.contains(self.pattern, regex=True)]
            result = cols.tolist()
        else:
            result = super(ColumnSelector, self).__call__(df)

        return result

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        attrs = []
        if self.pattern:
            attrs.append(f'pattern:{self.pattern}')
        if self.dtype_include:
            attrs.append(f'include:{self.dtype_include}')
        if self.dtype_exclude:
            attrs.append(f'exclude:{self.dtype_exclude}')

        s = f'{self.__class__.__name__}({", ".join(attrs)})'
        return s


class AutoCategoryColumnSelector(ColumnSelector):
    def __init__(self, pattern=None, *, dtype_include=None, dtype_exclude=None, cat_exponent=0.5):
        assert 0. < cat_exponent < 1.0

        super(AutoCategoryColumnSelector, self).__init__(pattern,
                                                         dtype_include=dtype_include,
                                                         dtype_exclude=dtype_exclude)

        self.cat_exponent = cat_exponent

    def __call__(self, df, *args, **kwargs):
        if self.pattern is not None or self.dtype_include is not None:
            selected = super().__call__(df)
        else:
            selected = []

        dtype_exclude = self.dtype_exclude if self.dtype_exclude is not None else []
        others = [c for c in df.columns.to_list() if c not in selected and str(df.dtypes[c]) not in dtype_exclude]
        if len(others) > 0:
            if isinstance(df, dd.DataFrame):
                nuniques = [df[c].nunique() for c in others]
                nuniques = {k: v for k, v in zip(others, dask.compute(*nuniques))}
                nunique_limit = dask.compute(df.shape[0])[0] ** self.cat_exponent
            else:
                nuniques = df[others].nunique(axis=0).to_dict()
                nunique_limit = df.shape[0] ** self.cat_exponent
            selected += [c for c, n in nuniques.items() if n <= nunique_limit]

        return selected


class TextColumnSelector(ColumnSelector):
    def __init__(self, pattern=None, *, dtype_include=None, dtype_exclude=None,
                 word_count_threshold=cfg.column_selector_text_word_count_threshold):
        assert isinstance(word_count_threshold, int) and word_count_threshold >= 1

        if dtype_include is None:
            dtype_include = ['object']

        super(TextColumnSelector, self).__init__(pattern,
                                                 dtype_include=dtype_include,
                                                 dtype_exclude=dtype_exclude)

        self.word_count_threshold = word_count_threshold

    def __call__(self, df, *args, **kwargs):
        selected = super().__call__(df)
        if len(selected) > 0 and self.word_count_threshold > 1:
            word_count = df[selected].applymap(self._word_count).max(axis=0)
            if isinstance(df, dd.DataFrame):
                word_count = word_count.compute()
            selected = [c for c, n in word_count.to_dict().items() if n >= self.word_count_threshold]

        return selected

    @staticmethod
    def _word_count(s):
        if not isinstance(s, str):
            return 0

        exist_chinese = False
        if _jieba_installed:
            for ch in s:
                if u'\u4e00' <= ch <= u'\u9fff':
                    exist_chinese = True
                    break

        if exist_chinese:
            word_count = 0
            for w in jieba.cut(s):
                if len(w.strip()) > 0:
                    word_count += 1
        else:
            word_count = len(s.split())

        return word_count


class LatLongColumnSelector:
    def __call__(self, df):
        cols = column_object(df)
        if cols is None or len(cols) < 1:
            return cols

        if isinstance(df, dd.DataFrame):
            row = df.reduction(LatLongColumnSelector._reduce_is_latlong,
                               aggregate=np.all, aggregate_kwargs=dict(axis=0),
                               meta={c: 'bool' for c in cols}
                               ).compute()
        elif isinstance(df, pd.DataFrame):
            row = LatLongColumnSelector._reduce_is_latlong(df)
        else:
            raise ValueError(f'Unsupported dataframe type "{type(df)}"')

        r = [k for k, v in row.to_dict().items() if v]
        return r

    @staticmethod
    def _is_latlong(v):
        try:
            return v is None or \
                   (isinstance(v, tuple) and len(v) == 2 and -90.0 <= v[0] <= 90.0 and -180.0 <= v[1] <= 180.0)
        except:
            return False

    @staticmethod
    def _reduce_is_latlong(df):
        fn = np.vectorize(LatLongColumnSelector._is_latlong, otypes=[np.bool], signature='()->()')
        return df.apply(fn).all(axis=0)


class MinMaxColumnSelector(object):
    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

    def __call__(self, df):
        if isinstance(df, dd.DataFrame):
            return self._select_dask_dataframe(df)
        elif isinstance(df, pd.DataFrame):
            return self._select_pandas_dataframe(df)
        else:
            raise ValueError(f'Unsupported dataframe type "{type(df)}"')

    def _select_pandas_dataframe(self, df):
        if self.min is not None and self.max is not None:
            df = df.aggregate(['min', 'max'])
            df = df.loc[:, (df.loc['min'] >= self.min) & (df.loc['max'] <= self.max)]
        elif self.min is not None:
            df = df.aggregate(['min'])
            df = df.loc[:, df.loc['min'] >= self.min]
        elif self.max is not None:
            df = df.aggregate(['max'])
            df = df.loc[:, df.loc['max'] <= self.max]

        return list(df.columns)

    def _select_dask_dataframe(self, df):
        if self.min is not None and self.max is not None:
            min_values, max_values = dask.compute(df.reduction(np.min, np.min),
                                                  df.reduction(np.max, np.max))
            df = pd.DataFrame({'min': min_values, 'max': max_values
                               }).T
            df = df.loc[:, (df.loc['min'] >= self.min) & (df.loc['max'] <= self.max)]
        elif self.min is not None:
            df = pd.DataFrame({'min': df.reduction(np.min, np.min).compute()
                               }).T
            df = df.loc[:, df.loc['min'] >= self.min]
        elif self.max is not None:
            df = pd.DataFrame({'max': df.reduction(np.max, np.max).compute()
                               }).T
            df = df.loc[:, df.loc['max'] <= self.max]

        return list(df.columns)


class CompositedColumnSelector(object):
    def __init__(self, selectors):
        assert isinstance(selectors, (tuple, list)) and len(selectors) > 0
        self.selectors = selectors

    def __call__(self, df):
        n = len(self.selectors)
        for i, selector in enumerate(self.selectors):
            columns = selector(df)
            if (i == n - 1) or len(columns) == 0:
                return columns

            df = df[columns]

        return list(df.columns)  # un-reached


column_all = ColumnSelector()
column_object_category_bool = ColumnSelector(dtype_include=['object', 'category', 'bool'])
column_object_category_bool_with_auto = AutoCategoryColumnSelector(dtype_include=['object', 'category', 'bool'],
                                                                   cat_exponent=0.5)
column_text = TextColumnSelector(dtype_include=['object'])
column_latlong = LatLongColumnSelector()

column_object = ColumnSelector(dtype_include=['object'])
column_category = ColumnSelector(dtype_include=['category'])
column_bool = ColumnSelector(dtype_include=['bool'])
column_number = ColumnSelector(dtype_include='number')
column_number_exclude_timedelta = ColumnSelector(dtype_include='number', dtype_exclude='timedelta')
column_object_category_bool_int = ColumnSelector(
    dtype_include=['object', 'category', 'bool', 'int16', 'int32', 'int64'])

column_timedelta = ColumnSelector(dtype_include='timedelta')
column_datetimetz = ColumnSelector(dtype_include='datetimetz')
column_datetime = ColumnSelector(dtype_include='datetime')
column_all_datetime = ColumnSelector(dtype_include=['datetime', 'datetimetz'])
column_int = ColumnSelector(dtype_include=['int16', 'int32', 'int64'])
column_exclude_datetime = ColumnSelector(
    dtype_exclude=['timedelta', 'datetime', 'datetimetz', 'period[M]', 'period[D]', 'period[Q]'])

column_zero_or_positive_int32 = CompositedColumnSelector(
    selectors=[column_int,
               MinMaxColumnSelector(0, np.iinfo(np.int32).max)]
)

column_positive_int32 = CompositedColumnSelector(
    selectors=[column_int,
               MinMaxColumnSelector(1, np.iinfo(np.int32).max)]
)


def column_min_max(X, min_value=None, max_value=None):
    selector = MinMaxColumnSelector(min_value, max_value)
    return selector(X)


def column_skewness_kurtosis(X, skew_threshold=0.5, kurtosis_threshold=0.5, columns=None):
    if columns is None:
        columns = column_number_exclude_timedelta(X)
    skew_values = skew(X[columns], axis=0, nan_policy='omit')
    kurtosis_values = kurtosis(X[columns], axis=0, nan_policy='omit')
    selected = [c for i, c in enumerate(columns) if
                abs(skew_values[i]) > skew_threshold or abs(kurtosis_values[i]) > kurtosis_threshold]
    return selected


def column_skewness_kurtosis_diff(X_1, X_2, diff_threshold=5, columns=None, smooth_fn=np.log, skewness_weights=1,
                                  kurtosis_weights=0):
    skew_x_1, skew_x_2, kurtosis_x_1, kurtosis_x_2, columns = calc_skewness_kurtosis(X_1, X_2, columns, smooth_fn)
    diff = np.log(
        abs(skew_x_1 - skew_x_2) * skewness_weights + np.log(abs(kurtosis_x_1 - kurtosis_x_2)) * kurtosis_weights)
    if isinstance(diff_threshold, tuple):
        index = np.argwhere((diff > diff_threshold[0]) & (diff <= diff_threshold[1]))
    else:
        index = np.argwhere(diff > diff_threshold)
    selected = [c for i, c in enumerate(columns) if i in index]
    return selected


def calc_skewness_kurtosis(X_1, X_2, columns=None, smooth_fn=np.log):
    if columns is None:
        columns = column_number_exclude_timedelta(X_1)
    X_1_t = X_1[columns]
    X_2_t = X_2[columns]
    if smooth_fn is not None:
        X_1_t[columns] = smooth_fn(X_1_t)
        X_2_t[columns] = smooth_fn(X_2_t)

    skew_x_1 = skew(X_1_t, axis=0, nan_policy='omit')
    skew_x_2 = skew(X_2_t, axis=0, nan_policy='omit')
    kurtosis_x_1 = kurtosis(X_1_t, axis=0, nan_policy='omit')
    kurtosis_x_2 = kurtosis(X_2_t, axis=0, nan_policy='omit')
    return skew_x_1, skew_x_2, kurtosis_x_1, kurtosis_x_2, columns



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from tpot import TPOTClassifier, TPOTRegressor
from . import BaseEstimator
from ..column_selector import column_object_category_bool
from ..sklearn_ex import SafeOrdinalEncoder


class TpotEstimator(BaseEstimator):
    def __init__(self, task, **kwargs):
        super(TpotEstimator, self).__init__(task)
        if task == 'regression':
            self.tpot = TPOTRegressor(**kwargs)
        else:
            self.tpot = TPOTClassifier(**kwargs)
        self.name = 'tpot'
        self.label_encoder = None
        self.obj_cols = None

    def train(self, X, y, X_test):
        self.obj_cols = column_object_category_bool(X)
        self.label_encoder = SafeOrdinalEncoder()
        X[self.obj_cols] = self.label_encoder.fit_transform(X[self.obj_cols])
        self.tpot.fit(X, y)

    def predict_proba(self, X):
        X[self.obj_cols] = self.label_encoder.transform(X[self.obj_cols])
        proba = self.tpot.predict_proba(X)
        print(f'proba.shape:{proba.shape}')
        return proba

    def predict(self, X):
        X[self.obj_cols] = self.label_encoder.transform(X[self.obj_cols])
        return self.tpot.predict(X)



import glob
import os
import sys
from traitlets import Unicode
from traitlets.config import Application, Configurable


class Configuration(Application):
    config_dir = Unicode('./conf',
                         help='The file system directory which contains all configuration files'
                         ).tag(config=True)

    aliases = {
        'log-level': 'Application.log_level',
        'config-dir': 'Configuration.config_dir',
    }

    def __init__(self, **kwargs):
        super(Configuration, self).__init__(**kwargs)

        aliases_keys = self.aliases.keys()
        argv = [a for a in sys.argv if any([a.startswith(f'--{k}') for k in aliases_keys])]
        super(Configuration, self).initialize(argv)

        if len(self.config_dir) > 0:
            config_dir = os.path.abspath(os.path.expanduser(self.config_dir))
            for f in glob.glob(f'{config_dir}/*.py', recursive=False):
                self.load_config_file(f)


_conf = Configuration()

generate_config_file = _conf.generate_config_file


def configure():
    """
    Annotation utility to configure one configurable class
    """

    def wrapper(c):
        assert issubclass(c, Configurable)
        o = c(parent=_conf)

        if c not in _conf.classes:
            _conf.classes += [c]

        return o

    return wrapper


def configure_and_observe(obj, observe_names, observe_handle):
    """
    Annotation utility to configure one configurable class and observe it (or other configured one)
    """
    assert (obj is None or isinstance(obj, Configurable)) \
           and callable(observe_handle) \
           and isinstance(observe_names, (tuple, list, str))

    def wrapper_and_observe(c):
        assert issubclass(c, Configurable)
        o = c(parent=_conf)

        if c not in _conf.classes:
            _conf.classes += [c]

        names = observe_names if isinstance(observe_names, (tuple, list)) else [observe_names]
        if obj is None:
            o.observe(observe_handle, names)
        else:
            obj.observe(observe_handle, names)

        return o

    return wrapper_and_observe


def observe(obj, names, handle):
    """
    A utility to observe configured object
    """
    assert isinstance(obj, Configurable) and callable(handle) \
           and isinstance(names, (tuple, list, str))

    names = names if isinstance(names, (tuple, list)) else [names]
    obj.observe(handle, names)

    return obj



# -*- coding:utf-8 -*-
"""

"""
import datetime
import json
import os
import time

import numpy as np
import pandas as pd
from IPython.display import display, update_display, display_markdown
from tqdm.auto import tqdm

from ..utils import logging, fs, to_repr

logger = logging.get_logger(__name__)


class Callback():
    def __init__(self):
        pass

    def on_search_start(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                        **fit_kwargs):
        pass

    def on_search_end(self, hyper_model):
        pass

    def on_search_error(self, hyper_model):
        pass

    def on_build_estimator(self, hyper_model, space, estimator, trial_no):
        pass

    def on_trial_begin(self, hyper_model, space, trial_no):
        pass

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        pass

    def on_trial_error(self, hyper_model, space, trial_no):
        pass

    def on_skip_trial(self, hyper_model, space, trial_no, reason, reward, improved, elapsed):
        pass

    def __repr__(self):
        return to_repr(self)


class EarlyStoppingError(RuntimeError):
    def __init__(self, *arg):
        self.args = arg


class EarlyStoppingCallback(Callback):
    REASON_TRIAL_LIMIT = 'max_no_improvement_trials'
    REASON_TIME_LIMIT = 'time_limit'
    REASON_EXPECTED_REWARD = 'expected_reward'

    def __init__(self, max_no_improvement_trials=0, mode='min', min_delta=0, time_limit=None, expected_reward=None):
        super(Callback, self).__init__()
        # assert time_limit is None or time_limit > 60, 'If `time_limit` is not None, it must be greater than 60.'

        # settings
        if mode == 'min':
            self.op = np.less
        elif mode == 'max':
            self.op = np.greater
        else:
            raise ValueError(f'Unsupported mode:{mode}')
        self.max_no_improvement_trials = max_no_improvement_trials
        self.mode = mode
        self.min_delta = min_delta
        self.time_limit = time_limit
        self.expected_reward = expected_reward

        # running state
        self.start_time = None
        self.best_reward = None
        self.best_trial_no = None
        self.counter_no_improvement_trials = 0
        self.triggered = None
        self.triggered_reason = None

    def on_search_start(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                        **fit_kwargs):
        self.triggered = False
        self.triggered_reason = None

    def on_trial_begin(self, hyper_model, space, trial_no):
        if self.start_time is None:
            self.start_time = time.time()

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        if self.start_time is None:
            self.start_time = time.time()

        time_total = time.time() - self.start_time

        if self.time_limit is not None and self.time_limit > 0:
            if time_total > self.time_limit:
                self.triggered = True
                self.triggered_reason = self.REASON_TIME_LIMIT

        if self.expected_reward is not None and self.expected_reward != 0.0:
            if self.op(reward, self.expected_reward):
                self.triggered = True
                self.triggered_reason = self.REASON_EXPECTED_REWARD

        if self.max_no_improvement_trials is not None and self.max_no_improvement_trials > 0:
            if self.best_reward is None:
                self.best_reward = reward
                self.best_trial_no = trial_no
            else:
                if self.op(reward, self.best_reward - self.min_delta):
                    self.best_reward = reward
                    self.best_trial_no = trial_no
                    self.counter_no_improvement_trials = 0
                else:
                    self.counter_no_improvement_trials += 1
                    if self.counter_no_improvement_trials >= self.max_no_improvement_trials:
                        self.triggered = True
                        self.triggered_reason = self.REASON_TRIAL_LIMIT

        if self.triggered:
            msg = f'Early stopping on trial : {trial_no}, reason: {self.triggered_reason}, ' \
                  f'best reward: {self.best_reward}, best trial: {self.best_trial_no}, ' \
                  f'elapsed seconds: {time_total}'
            if logger.is_info_enabled():
                logger.info(msg)
            raise EarlyStoppingError(msg)


class FileLoggingCallback(Callback):
    def __init__(self, searcher, output_dir=None):
        super(FileLoggingCallback, self).__init__()

        self.output_dir = self._prepare_output_dir(output_dir, searcher)

    @staticmethod
    def open(file_path, mode):
        return open(file_path, mode=mode)

    @staticmethod
    def mkdirs(dir_path, exist_ok):
        os.makedirs(dir_path, exist_ok=exist_ok)

    def _prepare_output_dir(self, log_dir, searcher):
        if log_dir is None:
            log_dir = 'log'
        if log_dir[-1] == '/':
            log_dir = log_dir[:-1]

        running_dir = f'exp_{searcher.__class__.__name__}_{datetime.datetime.now().__format__("%m%d-%H%M%S")}'
        output_path = os.path.expanduser(f'{log_dir}/{running_dir}')

        self.mkdirs(output_path, exist_ok=True)
        return output_path

    def on_build_estimator(self, hyper_model, space, estimator, trial_no):
        pass

    def on_trial_begin(self, hyper_model, space, trial_no):
        pass
        # with open(f'{self.output_dir}/trial_{trial_no}.log', 'w') as f:
        #     f.write(space.params_summary())

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        with self.open(f'{self.output_dir}/trial_{improved}_{trial_no:04d}_{reward:010.8f}_{elapsed:06.2f}.log',
                       'w') as f:
            f.write(space.params_summary())
            f.write('\r\n----------------Summary for Searcher----------------\r\n')
            f.write(hyper_model.searcher.summary())

        topn = 10
        diff = hyper_model.history.diff(hyper_model.history.get_top(topn))
        with self.open(f'{self.output_dir}/top_{topn}_diff.txt', 'w') as f:
            diff_str = json.dumps(diff, indent=5)
            f.write(diff_str)
            f.write('\r\n')
            f.write(hyper_model.searcher.summary())
        with self.open(f'{self.output_dir}/top_{topn}_config.txt', 'w') as f:
            trials = hyper_model.history.get_top(topn)
            configs = hyper_model.export_configuration(trials)
            for trial, conf in zip(trials, configs):
                f.write(f'Trial No: {trial.trial_no}, Reward: {trial.reward}\r\n')
                f.write(conf)
                f.write('\r\n---------------------------------------------------\r\n\r\n')

    def on_skip_trial(self, hyper_model, space, trial_no, reason, reward, improved, elapsed):
        with self.open(
                f'{self.output_dir}/trial_{reason}_{improved}_{trial_no:04d}_{reward:010.8f}_{elapsed:06.2f}.log',
                'w') as f:
            f.write(space.params_summary())

        topn = 5
        diff = hyper_model.history.diff(hyper_model.history.get_top(topn))
        with self.open(f'{self.output_dir}/top_{topn}_diff.txt', 'w') as f:
            diff_str = json.dumps(diff, indent=5)
            f.write(diff_str)


class FileStorageLoggingCallback(FileLoggingCallback):
    @staticmethod
    def open(file_path, mode):
        return fs.open(file_path, mode=mode)

    @staticmethod
    def mkdirs(dir_path, exist_ok):
        fs.mkdirs(dir_path, exist_ok=exist_ok)


class SummaryCallback(Callback):
    def __init__(self):
        super(SummaryCallback, self).__init__()

        self.start_search_time = None

    def on_search_start(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                        **fit_kwargs):
        self.start_search_time = time.time()

    def on_build_estimator(self, hyper_model, space, estimator, trial_no):
        # if logger.is_info_enabled():
        #     logger.info(f'\nTrial No:{trial_no}')
        #     logger.info(space.params_summary())
        estimator.summary()

    def on_trial_begin(self, hyper_model, space, trial_no):
        if logger.is_info_enabled():
            msg = f'\nTrial No:{trial_no}{space.params_summary()}\ntrial {trial_no} begin'
            logger.info(msg)

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        if logger.is_info_enabled():
            logger.info(f'trial end. reward:{reward}, improved:{improved}, elapsed:{elapsed}')
            logger.info(f'Total elapsed:{time.time() - self.start_search_time}')

    def on_skip_trial(self, hyper_model, space, trial_no, reason, reward, improved, elapsed):
        if logger.is_info_enabled():
            logger.info(f'&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&')
            logger.info(f'trial skip. reason:{reason},  reward:{reward}, improved:{improved}, elapsed:{elapsed}')
            logger.info(f'&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&')


class NotebookCallback(Callback):
    def __init__(self):
        super(NotebookCallback, self).__init__()

        self.current_trial_display_id = None
        self.search_summary_display_id = None
        self.best_trial_display_id = None
        self.title_display_id = None

        self.last_trial_no = 0
        self.last_reward = 0
        self.start_time = 0
        self.max_trials = 0

    def on_search_start(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                        **fit_kwargs):
        self.start_time = time.time()
        self.max_trials = max_trials

        df_holder = pd.DataFrame()
        settings = {'X': X.shape,
                    'y': y.shape,
                    'X_eval': X_eval.shape if X_eval is not None else None,
                    'y_eval': y_eval.shape if y_eval is not None else None,
                    'cv': cv,
                    'num_folds': num_folds,
                    'max_trials': max_trials,
                    # 'dataset_id': dataset_id,
                    # 'trail_store': trial_store,
                    'fit_kwargs': fit_kwargs.keys()
                    }
        df_settings = pd.DataFrame({k: [v] for k, v in settings.items()})

        display_markdown('#### Experiment Settings:', raw=True)
        display(hyper_model, display_id=False)
        display(df_settings, display_id=False)

        display_markdown('#### Trials Summary:', raw=True)
        handle = display(df_holder, display_id=True)
        if handle is not None:
            self.search_summary_display_id = handle.display_id

        display_markdown('#### Best Trial:', raw=True)
        handle = display(df_holder, display_id=True)
        if handle is not None:
            self.best_trial_display_id = handle.display_id

        handle = display({'text/markdown': '#### Current Trial:'}, raw=True, include=['text/markdown'],
                         display_id=True)
        if handle is not None:
            self.title_display_id = handle.display_id

        handle = display(df_holder, display_id=True)
        if handle is not None:
            self.current_trial_display_id = handle.display_id

    def on_trial_begin(self, hyper_model, space, trial_no):
        df_summary = pd.DataFrame([(trial_no, self.last_reward, hyper_model.best_trial_no,
                                    hyper_model.best_reward,
                                    time.time() - self.start_time,
                                    len([t for t in hyper_model.history.trials if t.succeeded]),
                                    self.max_trials)],
                                  columns=['Trial No.', 'Previous reward', 'Best trial', 'Best reward',
                                           'Total elapsed', 'Valid trials',
                                           'Max trials'])
        if self.search_summary_display_id is not None:
            update_display(df_summary, display_id=self.search_summary_display_id)

        if self.current_trial_display_id is not None:
            update_display(space, display_id=self.current_trial_display_id)

    def on_search_end(self, hyper_model):
        df_summary = pd.DataFrame([(self.last_trial_no, self.last_reward, hyper_model.best_trial_no,
                                    hyper_model.best_reward,
                                    time.time() - self.start_time,
                                    len([t for t in hyper_model.history.trials if t.succeeded]),
                                    self.max_trials)],
                                  columns=['Trial No.', 'Previous reward', 'Best trial', 'Best reward',
                                           'Total elapsed', 'Valid trials',
                                           'Max trials'])
        if self.search_summary_display_id is not None:
            update_display(df_summary, display_id=self.search_summary_display_id)

        if self.title_display_id is not None:
            update_display({'text/markdown': '#### Top trials:'}, raw=True, include=['text/markdown'],
                           display_id=self.title_display_id)

        df_best_trials = pd.DataFrame([
            (t.trial_no, t.reward, t.elapsed, t.space_sample.vectors) for t in hyper_model.get_top_trials(5)],
            columns=['Trial No.', 'Reward', 'Elapsed', 'Space Vector'])
        if self.current_trial_display_id is not None:
            update_display(df_best_trials, display_id=self.current_trial_display_id)

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        self.last_trial_no = trial_no
        self.last_reward = reward

        best_trial = hyper_model.get_best_trial()
        if best_trial is not None and self.best_trial_display_id is not None:
            update_display(best_trial.space_sample, display_id=self.best_trial_display_id)

    def on_trial_error(self, hyper_model, space, trial_no):
        self.last_trial_no = trial_no
        self.last_reward = 'ERR'


class ProgressiveCallback(Callback):
    def __init__(self):
        super(ProgressiveCallback, self).__init__()

        self.pbar = None

    def on_search_start(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                        **fit_kwargs):
        self.pbar = tqdm(total=max_trials, leave=False, desc='search')

    def on_search_end(self, hyper_model):
        self.pbar.update(self.pbar.total)
        self.pbar.close()
        self.pbar = None

    def on_search_error(self, hyper_model):
        self.on_search_end(hyper_model)

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        self.pbar.update(1)

    def on_trial_error(self, hyper_model, space, trial_no):
        self.pbar.update(1)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hypernets/dispatchers/predict/grpc/proto/predict.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hypernets/dispatchers/predict/grpc/proto/predict.proto',
  package='hypernets.dispatchers.predict.grpc.proto',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n6hypernets/dispatchers/predict/grpc/proto/predict.proto\x12(hypernets.dispatchers.predict.grpc.proto\"8\n\x0ePredictRequest\x12\x11\n\tdata_file\x18\x01 \x01(\t\x12\x13\n\x0bresult_file\x18\x02 \x01(\t\"X\n\x0fPredictResponse\x12\x11\n\tdata_file\x18\x01 \x01(\t\x12\x13\n\x0bresult_file\x18\x02 \x01(\t\x12\x0c\n\x04\x63ode\x18\x03 \x01(\x05\x12\x0f\n\x07message\x18\x04 \x01(\t2\x93\x01\n\x0ePredictService\x12\x80\x01\n\x07predict\x12\x38.hypernets.dispatchers.predict.grpc.proto.PredictRequest\x1a\x39.hypernets.dispatchers.predict.grpc.proto.PredictResponse\"\x00\x62\x06proto3')
)




_PREDICTREQUEST = _descriptor.Descriptor(
  name='PredictRequest',
  full_name='hypernets.dispatchers.predict.grpc.proto.PredictRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictRequest.data_file', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictRequest.result_file', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=100,
  serialized_end=156,
)


_PREDICTRESPONSE = _descriptor.Descriptor(
  name='PredictResponse',
  full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.data_file', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='result_file', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.result_file', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.code', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='hypernets.dispatchers.predict.grpc.proto.PredictResponse.message', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=246,
)

DESCRIPTOR.message_types_by_name['PredictRequest'] = _PREDICTREQUEST
DESCRIPTOR.message_types_by_name['PredictResponse'] = _PREDICTRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PredictRequest = _reflection.GeneratedProtocolMessageType('PredictRequest', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTREQUEST,
  '__module__' : 'hypernets.dispatchers.predict.grpc.proto.predict_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.predict.grpc.proto.PredictRequest)
  })
_sym_db.RegisterMessage(PredictRequest)

PredictResponse = _reflection.GeneratedProtocolMessageType('PredictResponse', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTRESPONSE,
  '__module__' : 'hypernets.dispatchers.predict.grpc.proto.predict_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.predict.grpc.proto.PredictResponse)
  })
_sym_db.RegisterMessage(PredictResponse)



_PREDICTSERVICE = _descriptor.ServiceDescriptor(
  name='PredictService',
  full_name='hypernets.dispatchers.predict.grpc.proto.PredictService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=249,
  serialized_end=396,
  methods=[
  _descriptor.MethodDescriptor(
    name='predict',
    full_name='hypernets.dispatchers.predict.grpc.proto.PredictService.predict',
    index=0,
    containing_service=None,
    input_type=_PREDICTREQUEST,
    output_type=_PREDICTRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PREDICTSERVICE)

DESCRIPTOR.services_by_name['PredictService'] = _PREDICTSERVICE

# @@protoc_insertion_point(module_scope)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from ._experiment import Experiment, ExperimentCallback
from .general import GeneralExperiment
from .compete import CompeteExperiment, SteppedExperiment, StepNames
from ._callback import ConsoleCallback, SimpleNotebookCallback
from ._maker import make_experiment



from hypernets.tabular import sklearn_ex as skex, dask_ex as dex
from hypernets.tabular.cache import cache, CacheCallback
from hypernets.tabular.datasets import dsutils
from hypernets.utils import hash_data, Counter


class CacheCounter(CacheCallback):
    def __init__(self):
        super(CacheCounter, self).__init__()

        self.enter_counter = Counter()
        self.apply_counter = Counter()
        self.store_counter = Counter()

    def on_enter(self, fn, *args, **kwargs):
        self.enter_counter()

    def on_apply(self, fn, cached_data, *args, **kwargs):
        self.apply_counter()

    def on_store(self, fn, cached_data, *args, **kwargs):
        self.store_counter()

    def reset(self):
        self.enter_counter.reset()
        self.apply_counter.reset()
        self.store_counter.reset()


class CachedMultiLabelEncoder(skex.MultiLabelEncoder):
    @cache(attr_keys='columns', attrs_to_restore='columns,encoders')
    def fit_transform(self, X, *args):
        return super().fit_transform(X, *args)

    @cache(attr_keys='columns', attrs_to_restore='columns,encoders')
    def fit_transform_as_tuple_result(self, X, *args):
        Xt = super().fit_transform(X.copy(), *args)
        return X, Xt


class CachedDaskMultiLabelEncoder(dex.MultiLabelEncoder):
    cache_counter = CacheCounter()

    @cache(attr_keys='columns',
           attrs_to_restore='columns,dtype,categorical_columns_,non_categorical_columns_,categories_',
           callbacks=cache_counter)
    def fit_transform(self, X, *args):
        return super().fit_transform(X, *args)

    @cache(attr_keys='columns',
           attrs_to_restore='columns,dtype,categorical_columns_,non_categorical_columns_,categories_',
           callbacks=cache_counter)
    def fit_transform_as_array(self, X, *args):
        X = super().fit_transform(X, *args)
        return X.to_dask_array(lengths=True)


def test_cache():
    df = dsutils.load_bank().head(10000)
    t = skex.MultiLabelEncoder()
    X = t.fit_transform(df.copy())

    t1 = CachedMultiLabelEncoder()
    X1 = t1.fit_transform(df.copy())
    t2 = CachedMultiLabelEncoder()
    X2 = t2.fit_transform(df.copy())
    assert hash_data(X) == hash_data(X1) == hash_data(X2)

    t3 = CachedMultiLabelEncoder()
    X3 = t3.fit_transform_as_tuple_result(df.copy())
    t4 = CachedMultiLabelEncoder()
    X4 = t4.fit_transform_as_tuple_result(df.copy())
    assert isinstance(X3, (tuple, list))
    assert isinstance(X4, (tuple, list))
    assert hash_data(X3[1]) == hash_data(X4[1])


def test_cache_dask():
    cache_counter = CachedDaskMultiLabelEncoder.cache_counter
    df = dex.dd.from_pandas(dsutils.load_bank().head(10000), npartitions=2)

    t = dex.MultiLabelEncoder()
    X = t.fit_transform(df.copy())

    cache_counter.reset()
    t1 = CachedDaskMultiLabelEncoder()
    X1 = t1.fit_transform(df.copy())
    t2 = CachedDaskMultiLabelEncoder()
    X2 = t2.fit_transform(df.copy())

    assert hash_data(X) == hash_data(X1) == hash_data(X2)
    assert cache_counter.enter_counter.value == 2
    assert cache_counter.apply_counter.value <= 2
    assert cache_counter.store_counter.value <= 2
    assert cache_counter.apply_counter.value + cache_counter.store_counter.value == 2

    cache_counter.reset()
    t3 = CachedDaskMultiLabelEncoder()
    X3 = t3.fit_transform_as_array(df.copy())
    t4 = CachedDaskMultiLabelEncoder()
    X4 = t4.fit_transform_as_array(df.copy())

    assert hash_data(X3) == hash_data(X4)
    assert cache_counter.enter_counter.value == 2
    assert cache_counter.apply_counter.value <= 2
    assert cache_counter.store_counter.value <= 2
    assert cache_counter.apply_counter.value + cache_counter.store_counter.value == 2



# -*- coding:utf-8 -*-
"""

"""
from ..core import TrialHistory
from ..core.callbacks import EarlyStoppingError
from ..core.searcher import Searcher, OptimizeDirection


class PlaybackSearcher(Searcher):
    def __init__(self, history: TrialHistory, top_n=None, reverse=False,
                 optimize_direction=OptimizeDirection.Minimize):
        assert history is not None
        assert len(history.trials) > 0

        self.history = history
        self.top_n = top_n if top_n is not None else len(history.trials)
        self.samples = [t.space_sample for t in self.history.get_top(self.top_n)]
        self.index = 0
        self.reverse = reverse

        if reverse:
            self.samples.reverse()

        super(PlaybackSearcher, self).__init__(None, use_meta_learner=False, optimize_direction=optimize_direction)

    @property
    def parallelizable(self):
        return True

    def sample(self):
        if self.index >= len(self.samples):
            raise EarlyStoppingError('no more samples.')
        sample = self.samples[self.index]
        self.index += 1
        return sample

    def update_result(self, space, result):
        pass



# -*- coding:utf-8 -*-
"""

"""
import copy

import lightgbm
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from hypernets.tabular import sklearn_ex as skex, dask_ex as dex
from hypernets.tabular.column_selector import column_object_category_bool, column_number_exclude_timedelta
from hypernets.tabular.dataframe_mapper import DataFrameMapper
from hypernets.utils import const


def general_preprocessor(X):
    if dex.is_dask_dataframe(X):
        import dask_ml.impute as dimp
        import dask_ml.preprocessing as dpre
        cat_steps = [('imputer_cat', dimp.SimpleImputer(strategy='constant', fill_value='')),
                     ('encoder', dex.SafeOrdinalEncoder())]
        num_steps = [('imputer_num', dimp.SimpleImputer(strategy='mean')),
                     ('scaler', dpre.StandardScaler())]
    else:
        cat_steps = [('imputer_cat', SimpleImputer(strategy='constant', fill_value='')),
                     ('encoder', skex.SafeOrdinalEncoder())]
        num_steps = [('imputer_num', SimpleImputer(strategy='mean')),
                     ('scaler', StandardScaler())]

    cat_transformer = Pipeline(steps=cat_steps)
    num_transformer = Pipeline(steps=num_steps)

    preprocessor = DataFrameMapper(features=[(column_object_category_bool, cat_transformer),
                                             (column_number_exclude_timedelta, num_transformer)],
                                   input_df=True,
                                   df_out=True)
    return preprocessor


def _wrap_predict_proba(estimator):
    orig_predict_proba = estimator.predict_proba

    def __predict_proba(*args, **kwargs):
        proba = orig_predict_proba(*args, **kwargs)
        proba = dex.fix_binary_predict_proba_result(proba)
        return proba

    setattr(estimator, '_orig_predict_proba', orig_predict_proba)
    setattr(estimator, 'predict_proba', __predict_proba)
    return estimator


def general_estimator(X, estimator=None, task=None):
    def default_gbm():
        cls = lightgbm.LGBMRegressor if task == const.TASK_REGRESSION else lightgbm.LGBMClassifier
        return cls(n_estimators=50,
                   num_leaves=15,
                   max_depth=5,
                   subsample=0.5,
                   subsample_freq=1,
                   colsample_bytree=0.8,
                   reg_alpha=1,
                   reg_lambda=1,
                   importance_type='gain', )

    def default_dt():
        cls = DecisionTreeRegressor if task == const.TASK_REGRESSION else DecisionTreeClassifier
        return cls(min_samples_leaf=20, min_impurity_decrease=0.01)

    def default_rf():
        cls = RandomForestRegressor if task == const.TASK_REGRESSION else RandomForestClassifier
        return cls(min_samples_leaf=20, min_impurity_decrease=0.01)

    def default_dask_xgb():
        import dask_xgboost
        cls = dask_xgboost.XGBRegressor if task == const.TASK_REGRESSION else dask_xgboost.XGBClassifier
        return cls(max_depth=5,
                   n_estimators=50,
                   min_child_weight=3,
                   gamma=1,
                   # reg_alpha=0.1,
                   # reg_lambda=0.1,
                   subsample=0.6,
                   colsample_bytree=0.6,
                   eval_metric='auc',
                   # objective='binary:logitraw',
                   tree_method='approx', )

    if dex.is_dask_dataframe(X):
        try:
            estimator_ = default_dask_xgb()
            estimator_ = _wrap_predict_proba(estimator_)
        except ImportError:  # failed to import dask_xgboost
            estimator_ = default_gbm()
            estimator_ = dex.wrap_local_estimator(estimator_)
    else:
        if estimator is None or estimator == 'gbm':
            estimator_ = default_gbm()
        elif estimator == 'dt':
            estimator_ = default_dt()
        elif estimator == 'rf':
            estimator_ = default_rf()
        else:
            estimator_ = copy.deepcopy(estimator)

    return estimator_
